
<h3>Related Topics</h3>
<ul>
<li><a href="{{ pathto(master_doc) }}">Documentation overview</a></li>
{%- for parent in parents %}
<li><a href="{{ parent.link|e }}">{{ parent.title }}</a><li>
</ul>
{%- endfor %}
{%- if prev %}
  <p>
  <b>Previous</b><br/>
  <a href="{{ prev.link|e }}" title="{{ _('previous chapter') }}">{{ prev.title }}</a>
  </p>
{%- endif %}
{%- if next %}
  <p>
  <b>Next</b><br/>
  <a href="{{ next.link|e }}" title="{{ _('next chapter') }}">{{ next.title }}</a>
  </p>
{%- endif %}
